"""
This is uses the k-nearest-neighbors (KNN) algorithm for face recognition.

When should I use this example?
This example is useful when you wish to recognize a large set of known people,
and make a prediction for an unknown person in a feasible computation time.

Algorithm Description:
The knn classifier is first trained on a set of labeled (known) faces and can then predict the person
in an unknown image by finding the k most similar faces (images with closet face-features under euclidean distance)
in its training set, and performing a majority vote (possibly weighted) on their label.

For example, if k=3, and the three closest face images to the given image in the training set are one image of Biden
and two images of Obama, The result would be 'Obama'.

* This implementation uses a weighted vote, such that the votes of closer-neighbors are weighted more heavily.

Usage:

1. Prepare a set of images of the known people you want to recognize. Organize the images in a single directory
   with a sub-directory for each known person.

2. Then, call the 'train' function with the appropriate parameters. Make sure to pass in the 'model_save_path' if you
   want to save the model to disk so you can re-use the model without having to re-train it.

3. Call 'predict' and pass in your trained model to recognize the people in an unknown image.

NOTE: This example requires scikit-learn to be installed! You can install it with pip:

$ pip3 install scikit-learn

"""
import time

import cv2
import pkg_resources

from dronebuddylib.atoms.facerecognition import IFaceRecognition
from dronebuddylib.atoms.facerecognition.face_recognition_result import RecognizedFaceObject, RecognizedFaces
from dronebuddylib.models import AtomicEngineConfigurations

from dronebuddylib.utils.logger import Logger

import math

import cv2
from sklearn import neighbors, preprocessing
import os
import os.path
import pickle
import face_recognition
from face_recognition.face_recognition_cli import image_files_in_folder
from sklearn.metrics import accuracy_score
import joblib

from dronebuddylib.utils.utils import overwrite_file, read_from_file

logger = Logger()


class FaceRecognitionKNNImpl(IFaceRecognition):
    def recognize_face(self, image) -> RecognizedFaces:
        # Find face locations and encodings in the image
        X_face_locations = face_recognition.face_locations(image)
        faces_encodings = face_recognition.face_encodings(image, known_face_locations=X_face_locations)
        distance_threshold = 0.5
        # Ensure at least one face is found
        if len(faces_encodings) == 0:
            return []
        classifier_index = read_from_file(
            pkg_resources.resource_filename(__name__, "resources/models/classifiers/classifier_index.txt"))

        model_path = pkg_resources.resource_filename(__name__, "resources/models/classifiers/trained_knn_model" + str(
            classifier_index) + ".clf")
        # Ensure at least one face is found
        if len(faces_encodings) == 0:
            return []
        with open(model_path, 'rb') as f:
            model = pickle.load(f)

        # Use the KNN model to find the best matches for the test faces
        closest_distances = model.kneighbors(faces_encodings, n_neighbors=1)
        are_matches = [closest_distances[0][i][0] <= distance_threshold for i in range(len(X_face_locations))]

        # Predict classes and remove classifications that aren't within the threshold
        predictions = [(pred, loc) if rec else ("unknown", loc) for pred, loc, rec in
                       zip(model.predict(faces_encodings), X_face_locations, are_matches)]
        result = []
        for name, (top, right, bottom, left) in predictions:
            result.append(RecognizedFaceObject(name, [top, right, bottom, left]))
            print("- Found {} at ({}, {})".format(name, left, top))
        return RecognizedFaces(result)
    def remember_face(self, image=None, name=None) -> bool:
        face_classifier = cv2.CascadeClassifier(cv2.data.haarcascades + 'haarcascade_frontalface_default.xml')

        logger.log_info(self.get_class_name(), 'Place your face in front of the camera to create a data set.')
        logger.log_info(self.get_class_name(), 'If you see the message "Face not found, move a bit.", move a bit.')

        for i in range(5):
            logger.log_info(self.get_class_name(), 'Starting in ' + str(3 - i) + ' seconds...')
            time.sleep(1)

        logger.log_info(self.get_class_name(), 'Starting to create training data set for ' + name)
        self.create_data_set(face_classifier, name, 0, 900)
        logger.log_info(self.get_class_name(), 'Successfully created training data set for ' + name)
        logger.log_info(self.get_class_name(), 'Get ready for the test set creation ' + name)

        for i in range(5):
            logger.log_info(self.get_class_name(), 'Starting test set creation in ' + str(5 - i) + ' seconds...')
            time.sleep(1)

        logger.log_info(self.get_class_name(), 'Starting to create testing data set for ' + name)
        self.create_data_set(face_classifier, name, 1, 100)
        logger.log_info(self.get_class_name(), 'Successfully created testing data set for ' + name)

        logger.log_info(self.get_class_name(), 'Get ready for the test set creation ' + name)

        for i in range(5):
            logger.log_info(self.get_class_name(), 'Starting validation set creation in ' + str(5 - i) + ' seconds...')
            time.sleep(1)

        logger.log_info(self.get_class_name(), 'Starting to create validation data set for ' + name)

        self.create_data_set(face_classifier, name, 2, 50)
        logger.log_info(self.get_class_name(), 'Successfully created validation data set for ' + name)

        return True

    def train(self, train_dir, model_save_path=None, n_neighbors=None, knn_algo='ball_tree', verbose=False):
        """
        Trains a k-nearest neighbors classifier for face recognition.

        :param train_dir: directory that contains a sub-directory for each known person, with its name.

         (View in source code to see train_dir example tree structure)

         Structure:
            <train_dir>/
            ├── <person1>/
            │   ├── <somename1>.jpeg
            │   ├── <somename2>.jpeg
            │   ├── ...
            ├── <person2>/
            │   ├── <somename1>.jpeg
            │   └── <somename2>.jpeg
            └── ...

        :param model_save_path: (optional) path to save model on disk
        :param n_neighbors: (optional) number of neighbors to weigh in classification. Chosen automatically if not specified
        :param knn_algo: (optional) underlying data structure to support knn.default is ball_tree
        :param verbose: verbosity of training
        :return: returns knn classifier that was trained on the given data.
        """
        X = []
        y = []

        # Loop through each person in the training set
        for class_dir in os.listdir(train_dir):
            if not os.path.isdir(os.path.join(train_dir, class_dir)):
                continue

            # Loop through each training image for the current person
            for img_path in image_files_in_folder(os.path.join(train_dir, class_dir)):
                image = face_recognition.load_image_file(img_path)
                face_bounding_boxes = face_recognition.face_locations(image)

                if len(face_bounding_boxes) != 1:
                    # If there are no people (or too many people) in a training image, skip the image.
                    if verbose:
                        logger.log_debug(self.get_class_name(),
                                         "Image {} not suitable for training: {}"
                                         .format(img_path,
                                                 "Didn't find a face"
                                                 if len(face_bounding_boxes) < 1 else "Found more than one face"))
                else:
                    # Add face encoding for current image to the training set
                    X.append(face_recognition.face_encodings(image, known_face_locations=face_bounding_boxes)[0])
                    y.append(class_dir)

        # Determine how many neighbors to use for weighting in the KNN classifier
        if n_neighbors is None:
            n_neighbors = int(round(math.sqrt(len(X))))
            if verbose:
                logger.log_debug(self.get_class_name(), ("Chose n_neighbors automatically:" + str(n_neighbors)))

        # Create and train the KNN classifier
        knn_clf = neighbors.KNeighborsClassifier(n_neighbors=n_neighbors, algorithm=knn_algo, weights='distance')
        knn_clf.fit(X, y)

        # Save the trained KNN classifier
        if model_save_path is not None:
            with open(model_save_path, 'wb') as f:
                pickle.dump(knn_clf, f)

        return knn_clf

    def create_memory(self):
        logger.log_info(self.get_class_name(), 'Starting to create memory. Training KNN classifier')

        file_path = pkg_resources.resource_filename(__name__, "resources/models/classifiers/classifier_index.txt")
        classifier_index_read = read_from_file(file_path)
        classifier_index = int(classifier_index_read) + 1
        overwrite_file(file_path, str(classifier_index))

        classifier_path_name = pkg_resources.resource_filename(__name__,
                                                               "resources/models/classifiers/" + "trained_knn_model"
                                                               + str(classifier_index) + ".clf")
        training_path_name = pkg_resources.resource_filename(__name__, "resources/test_data/training_data")
        self.train(training_path_name,
                   model_save_path=classifier_path_name, n_neighbors=2,
                   verbose=True)
        logger.log_info(self.get_class_name(),
                        'Successfully created memory. KNN classifier trained successfully. Model saved at: '
                        + classifier_path_name)

        logger.log_debug(self.get_class_name(), 'Starting to test the classifier')
        accuracy_score = self.test_classifier()
        return accuracy_score

    def test_classifier(self):
        file_path = pkg_resources.resource_filename(__name__, "resources/models/classifiers/classifier_index.txt")

        classifier_index_read = read_from_file(file_path)
        classifier_index = int(classifier_index_read)
        classifier_path_name = pkg_resources.resource_filename(__name__,
                                                               "resources/models/classifiers/" + "trained_knn_model"
                                                               + str(classifier_index) + ".clf")
        testing_path_name = pkg_resources.resource_filename(__name__, "resources/test_data/testing_data")
        # Load the trained classifier
        classifier = joblib.load(classifier_path_name)

        # Load your test data
        features, true_labels, label_names = self.load_test_data(testing_path_name)

        # Predict using the classifier
        predicted_labels = classifier.predict(features)

        # Calculate and print the accuracy
        accuracy = accuracy_score(true_labels, predicted_labels)
        logger.log_info(self.get_class_name(), f"Accuracy: {accuracy * 100:.2f}%")
        return accuracy

    def extract_features(self, image_path):
        """
           Extract the face encodings from the given image path.
           """
        # Load the image file
        image = face_recognition.load_image_file(image_path)

        # Find face locations in the image
        face_locations = face_recognition.face_locations(image)

        # If no faces are found, return None
        if len(face_locations) == 0:
            return None

        # Assuming we'll use only the first face found in the image
        face_encoding = face_recognition.face_encodings(image, known_face_locations=face_locations)[0]

        return face_encoding

    def load_test_data(self, test_dir):
        """
           Load images and true labels from the test directory.
           """
        features = []
        labels = []
        label_names = []

        # Iterate over each directory in test_dir
        for label_name in os.listdir(test_dir):
            label_dir = os.path.join(test_dir, label_name)
            if not os.path.isdir(label_dir):
                continue  # Skip non-directory files

            # Store the name of the label
            label_names.append(label_name)

            # Iterate over each image in the directory
            for image_name in os.listdir(label_dir):
                image_path = os.path.join(label_dir, image_name)

                # Extract features from the image (i.e., face encodings)
                face_encoding = self.extract_features(image_path)
                if face_encoding is not None:
                    features.append(face_encoding)
                    labels.append(label_name)

        # Convert labels to a numerical format
        le = preprocessing.LabelEncoder()
        labels_encoded = le.fit_transform(labels)

        return features, labels, le.classes_

    # Load functions
    def face_extractor(self, img, face_classifier=None):
        # Load HAAR face classifier
        # Function detects faces and returns the cropped face
        # If no face detected, it returns the input image

        global cropped_face
        gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
        faces = face_classifier.detectMultiScale(gray, 1.3, 5)

        if faces is ():
            return None

        # Crop all faces found
        for (x, y, w, h) in faces:
            cropped_face = img[y:y + h, x:x + w]

        return cropped_face

    def create_data_set(self, face_classifier, participant_name, data_mode, data_set_size=900):

        cap = cv2.VideoCapture(0)
        count = 00

        if data_mode == 0:
            path_name = pkg_resources.resource_filename(__name__,
                                                        "resources/test_data/training_data/" + participant_name)
        elif data_mode == 1:
            path_name = pkg_resources.resource_filename(__name__,
                                                        "resources/test_data/testing_data/" + participant_name)
        elif data_mode == 2:
            path_name = pkg_resources.resource_filename(__name__,
                                                        "resources/test_data/validation_data/" + participant_name)
        else:
            path_name = pkg_resources.resource_filename(__name__,
                                                        "resources/test_data/training_data/" + participant_name)
        if not os.path.exists(path_name):
            os.makedirs(path_name)
        # Collect 100 samples of your face from webcam input
        while True:
            ret, frame = cap.read()
            if self.face_extractor(frame, face_classifier=face_classifier) is not None:
                count += 1
                face = cv2.resize(self.face_extractor(frame, face_classifier=face_classifier), (300, 300))

                # Save file in specified directory with unique name
                file_name_path = (path_name + "\\" + participant_name + "_" + str(count) + ".jpg")
                cv2.imwrite(file_name_path, face)
                # Put count on images and display live count
                cv2.putText(face, str(count), (50, 50), cv2.FONT_HERSHEY_COMPLEX, 1, (0, 255, 0), 2)
                cv2.imshow("FaceCropper", face)

            else:
                logger.log_warning(self.get_class_name(), 'Face not found, move a bit.')
                pass

            if cv2.waitKey(1) == 27 or count == data_set_size:  # 27 is the Esc Key
                break

        cap.release()
        cv2.destroyAllWindows()
        logger.log_warning(self.get_class_name(),
                           'Data set created successfully. ' + str(data_set_size) +
                           ' data collected. Data set can be found at the path: ' + path_name)

    def get_required_params(self) -> list:
        """
        Get the required parameters for the FaceRecognitionImpl class.

        Returns:
            list: A list of required parameters.
        """
        pass

    def get_optional_params(self) -> list:
        """
        Get the optional parameters for the FaceRecognitionImpl class.

        Returns:
            list: A list of optional parameters.
        """
        return [AtomicEngineConfigurations.FACE_RECOGNITION_ALGORITHM_NEIGHBORS,
                AtomicEngineConfigurations.FACE_RECOGNITION_ALGORITHM_NAME,
                AtomicEngineConfigurations.FACE_RECOGNITION_MODEL_SAVING_PATH,
                AtomicEngineConfigurations.FACE_RECOGNITION_TESTING_DATA_SET_SIZE,
                AtomicEngineConfigurations.FACE_RECOGNITION_TESTING_DATA_SET_SIZE,
                AtomicEngineConfigurations.FACE_RECOGNITION_VALIDATION_DATA_SET_SIZE, ]

    def get_class_name(self) -> str:
        """
        Get the class name of the FaceRecognitionImpl class.

        Returns:
            str: The class name.
        """
        return 'FACE_RECOGNITION'

    def get_algorithm_name(self) -> str:
        """
        Get the algorithm name of the FaceRecognitionImpl class.

        Returns:
            str: The algorithm name.
        """
        return 'Face Recognition'
